library( ANTsR )
library( ANTsRNet )
library( keras )
library( tensorflow )

dbseg <- function( x, unetfn, lowerq=0.001, upperq=0.95, referenceIntensity ) {
  templateP = Sys.glob( "./SPM/*" )
  template = antsImageRead( templateP[4] )
  templateP = templateP[1:3]
  msk = thresholdImage( x, 0.1*mean(x), Inf )
  img = cropImage( x * msk, msk ) %>% padImageByFactor(8)
  if ( ! missing( referenceIntensity ) )
    img = histogramMatchImage( img, referenceIntensity, 32, 8 ,useThresholdAtMeanIntensity = T )
  msk = cropImage( msk, msk ) %>% padImageByFactor(8)
  img = brainAgeR::standardizeIntensity( img, quantiles = c( lowerq, upperq ) )
  reg = antsRegistration( template, iMath(img,"Normalize"), "SyN" )
  priors = imageFileNames2ImageList( templateP )
  pX = array( dim = c( 1, dim( img ), 3 ) )
  for ( k in 1:length(priors) ) {
    priors[[k]] = antsApplyTransforms( img, priors[[k]], reg$invtransforms )
    pX[1,,,,k] = as.array( priors[[k]] )
    }
  unet = load_model_hdf5( unetfn )
  splitUp = tf$split( unet$output[[2]], as.integer(3), axis=as.integer(4) )
  unetP = keras_model( unet$inputs[[1]],  splitUp )
  # make it an array
  temp = tf$cast( array( as.array( img ) , dim = c( 1, dim(img), 1 ) ), 'float32' )
  with(tf$device("/cpu:0"), {
    if ( length( unetP$inputs ) == 2 ) newsegs = predict( unetP, list( temp, pX) )
    if ( length( unetP$inputs ) == 1 ) newsegs = predict( unetP, temp )
    TV <- tf$constant( 0 )
    for ( kk in 1:length(newsegs) )
        TV <- TV +
          tf$reduce_mean( tf$image$total_variation(  tf$cast(newsegs[[kk]][1,,,,1], 'float32') ) ) * 5e-5
      })
  for ( k in 1:length( newsegs ) ) {
    newsegs[[k]] = as.antsImage( newsegs[[k]][1,,,,1] ) %>% antsCopyImageInfo2( img )
    }
  mat = imageListToMatrix( newsegs[1:3], msk )
  segger = makeImage( msk, apply( mat, FUN=which.max, MARGIN=2) )
  return( list( img=img, segmentation = segger, probabilities = newsegs, TV=as.numeric(TV) ) )
  }

# model location ~/code/tissueSegmentation/travelingSubjects/models
mdlfnOR = "models/kmeansUnetDepth2NC1Bayes3ResORcsf20_all_sites_traveling.h5"
mdlfnSR = "models/kmeansUnetDepth2NC1Bayes3ResSRcsf20_all_sites_traveling.h5"
mdlfn = mdlfnSR # use SR for super resolution, OR for native ... this is a bit of a research question though
fn = "image.nii.gz"
img = antsImageRead( fn )
img = n3BiasFieldCorrection( img, 4 )
bxt = brainExtraction( img, "t1combined" )
bxtt = thresholdImage( bxt, 2, 3 ) # ants bxt
imgb = n4BiasFieldCorrection( img * bxtt, bxtt, 4 )
segs = dbseg( imgb, mdlfn )
fakeimg <- function( tarimg, segmentation, probs ) {
  m1 = mean( tarimg[ segmentation == 1 ] )
  m2 = mean( tarimg[ segmentation == 2 ] )
  m3 = mean( tarimg[ segmentation == 3 ] )
  m1 * probs[[1]] + m2 * probs[[2]] + m3 * probs[[3]]
}

faker = fakeimg( segs[[1]], segs[[2]], segs[[3]] )
imageError = mean( abs( faker - segs[[1]] ) )
volumes = labelGeometryMeasures( segs[[2]] )$Volume
# outputs should be:
## images = segs[[2]] ( the segmentation )
#           segs[[2]] ( the probabilities )
#  to csv file =  segs$TV, imageError and volumes
# possibly the plot below
segdf = data.frame( imageFileName=fn, segModel=mdlfn, CSFVol =  volumes[1], GMVol = volumes[2],  WMVol =  volumes[3], reconstructionError = imageError, totalVariation = segs$TV )
layout(matrix(1:2,nrow=1))
plot( segs[[1]], axis=3 )
plot( segs[[1]], segs[[2]], axis=3 )
write.csv( segdf, 'segmentation_summary_data.csv', row.names=F )
antsImageWrite( segs[[2]], 'segmentation_result.nii.gz' )
